/*
 * Copyright (C) 1994, Silicon Graphics, Inc.
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Silicon Graphics, Inc.
 *
 * RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
 * and Computer Software clause at DFARS 252.227-7013, and/or in similar or
 * successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
 * rights reserved under the Copyright Laws of the United States.
 */
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/prctl.h>
#include <sys/wait.h>
#include <unistd.h>
#include <fcntl.h>
#include <malloc.h>
#include <math.h>
#include <string.h>
#include <signal.h>
#include <audio.h>
#include <audiofile.h>
#include "RecordClass.h"


static void quietAFerror(long, const char *);
static void quietALerror(long, const char *, ...);

#define FALSE 0
#define TRUE 1

RecordClass::RecordClass(char *fname, char *mname)
{
#ifdef DEBUG
    fprintf(stderr, "RecordClass(%s, %s)\n", fname, mname);
#endif
    filename = fname;
    myname = mname;
    set_defaults();
    isLive = FALSE;
}


RecordClass::~RecordClass()
{
#ifdef DEBUG
    fprintf(stderr, "~RecordClass()\n");
#endif
}


/*
 * R E C O R D _ P A R A M S _ D E F A U L T S
 */
void
RecordClass::set_defaults()
{
#ifdef DEBUG
    fprintf(stderr, "set_defaults\n");
#endif
    long pvbuf[2];
    long pvbuflen;

    fork = 0; // do not fork process
    cilmi = 0; // do not kill process yet
    filefmt = AF_FILE_AIFF;

    time_limit_specified = 0;
    time_limit = 0.0;

    samps_per_frame = 2;
    bits_per_samp   = 16;

    pvbuf[0] = AL_INPUT_RATE;
    pvbuflen = 2;

    ALgetparams(AL_DEFAULT_DEVICE, pvbuf, pvbuflen);

    switch (pvbuf[1])
    {
        case AL_RATE_8000:  frames_per_sec = 8000; break;
        case AL_RATE_11025: frames_per_sec = 11025; break;
        case AL_RATE_16000: frames_per_sec = 16000; break;
        case AL_RATE_22050: frames_per_sec = 22050; break;
        case AL_RATE_32000: frames_per_sec = 32000; break;
        case AL_RATE_44100: frames_per_sec = 44100; break;
        case AL_RATE_48000: frames_per_sec = 48000; break;
        default:
            fprintf(stderr, "%s: audio processor reported bad input rate %d\n",
                                       myname, pvbuf[1]);
            exit(1);
    }
#ifdef DEBUG
  fprintf(stderr, "Set Defaults: frames_per_sec rate  = %d\n", frames_per_sec);
#endif

    compression = AF_COMPRESSION_NONE;
    verbose = 0;
}

void
RecordClass::setFrames(long t)
{

#ifdef DEBUG
    fprintf(stderr, "setFrames(%d)\n", t);
#endif

    switch(t)
      {
        case 48000:
        case 44100:
        case 32000:
        case 22050:
        case 16000:
        case 11025:
        case 8000:
	  frames_per_sec = t;
	  break;
        default:
	  fprintf(stderr, "%s: invalid sample rate %d\n",
		  myname, t);
       }
}


void RecordClass::setBits(long t)
{
#ifdef DEBUG
    fprintf(stderr, "setBits(%d)\n", t);
#endif
    switch(t)
      {
        case 8:
        case 16:
        case 24:
	  bits_per_samp = t;
	  break;
        default:
	  fprintf(stderr, "%s: invalid sample width %d\n",
		  myname, t);
      }
}

void RecordClass::setSamps(long t)
{
#ifdef DEBUG
    fprintf(stderr, "setSamps(%d)\n", t);
#endif
    switch(t) {
	case 1:
	case 2:
	  samps_per_frame = t;
	  break;
        default:
	  fprintf(stderr, "%s: invalid number of channels %d\n",
		  myname, t);
    }
}

void RecordClass::setTimeLimit(float t)
{
#ifdef DEBUG
    fprintf(stderr, "setTimeLimit(%d)\n", t);
#endif
    if (t > 0) {
	time_limit_specified = 1;
	time_limit = (double) t;
    } else {
	fprintf(stderr, "%s: invalid time limit %f seconds\n",
		myname, t);
    }
}

void RecordClass::setCompression(char *s)
{
#ifdef DEBUG
    fprintf(stderr, "setCompression(%s)\n", s);
#endif
    if (!strcmp(s, "g722"))
      {
	  compression = AF_COMPRESSION_G722;
      }
    else if (!strcmp(s, "ulaw"))
      {
	  compression = AF_COMPRESSION_G711_ULAW;
      }
    else if (!strcmp(s, "alaw"))
      {
	  compression = AF_COMPRESSION_G711_ALAW;
      }
    else
      {
	  fprintf(stderr, "%s: invalid compression %s\n",
		  myname, s);
      }
}

void RecordClass::setFormat(char *s)
{
#ifdef DEBUG
    fprintf(stderr, "setFormat(%s)\n", s);
#endif
    if  (!strcasecmp(s, "aiff"))
      {
	  filefmt = AF_FILE_AIFF;
      }
    else if (!strcasecmp(s, "aifc")
	     || !strcasecmp(s, "aiff-c")
	     || !strcasecmp(s, "aiffc"))
      {
	  filefmt = AF_FILE_AIFFC;
      }
    else
      {
	  fprintf(stderr, "%s: invalid file format %s\n", myname, s);
      }
}


void RecordClass::setFilename(char *s)
{
#ifdef DEBUG
    fprintf(stderr, "setFilename(%s)\n", s);
#endif
    filename = s;
}

void RecordClass::setVerbose(short i)
{
#ifdef DEBUG
    fprintf(stderr, "setVerbose(%d)\n", i);
#endif
    verbose = i;
}

void RecordClass::setQuiet(short i)
{
#ifdef DEBUG
    fprintf(stderr, "setQuiet(%d)\n", i);
#endif
    quiet = i;
}

// put this in begin recording
int
RecordClass::beginRecording()
{
#ifdef DEBUG
    fprintf(stderr, "beingRecording()\n");
#endif

    int n;
    ALport audio_port;
    ALconfig audio_port_config;
    AFfilehandle audio_file;
    AFfilesetup audio_file_setup;
    int fd;
    long pvlen;
    long pvbuf[2];

    isLive = TRUE;

    // change hardware sampling rate
    pvbuf[0] = AL_INPUT_RATE;
    pvlen = 2;
#ifdef DEBUG
  fprintf(stderr, "begin recording: frames_per_sec: %d\n", frames_per_sec);
#endif

    switch (frames_per_sec)
      {
        case 8000:  pvbuf[1] = AL_RATE_8000;  break;
        case 11025: pvbuf[1] = AL_RATE_11025;  break;
        case 16000: pvbuf[1] = AL_RATE_16000; break;
        case 22050: pvbuf[1] = AL_RATE_22050; break;
        case 32000: pvbuf[1] = AL_RATE_32000; break;
        case 44100: pvbuf[1] = AL_RATE_44100; break;
        default:
        case 48000: pvbuf[1] = AL_RATE_48000; break;
      }

    ALsetparams(AL_DEFAULT_DEVICE, pvbuf, pvlen);

    /*
     * check record parameters for consistency
     */
    if ((compression != AF_COMPRESSION_NONE)
          && (filefmt == AF_FILE_AIFF))
    {
        fprintf(stderr,
       "%s: can't store compressed audio data in an AIFF file: using AIFF-C\n",
            myname);
	filefmt = AF_FILE_AIFFC;
    }
    if ((bits_per_samp != 16)
          && (compression != AF_COMPRESSION_NONE))
    {
        fprintf(stderr,
         "%s: G.711 and G.722 encoders expect 16-bit sample data\n",
            myname);
	bits_per_samp = 16;
    }

    if (quiet)
    {
        ALseterrorhandler(quietALerror);
        AFseterrorhandler(quietAFerror);
    }

    /*
     * obtain a unix file descriptor for the input file
     */
    fd = open(filename, O_CREAT|O_TRUNC|O_WRONLY, 0644);
    if (fd < 0) {
       fprintf(stderr, "%s: could not create output file %s\n",
                   myname, filename);
       return -1;
    }

    /*
     * initialize the audio port and audio file structures
     */
    audio_port_config = ALnewconfig();
    audio_file_setup  = AFnewfilesetup();

    AFinitfilefmt(audio_file_setup, filefmt);
    AFinitcompression(audio_file_setup, AF_DEFAULT_TRACK, compression);

    switch (samps_per_frame) {
        case 1:
            ALsetchannels(audio_port_config, AL_MONO);
            AFinitchannels(audio_file_setup, AF_DEFAULT_TRACK, 1);
            break;
        default:
        case 2:
            ALsetchannels(audio_port_config, AL_STEREO);
            AFinitchannels(audio_file_setup, AF_DEFAULT_TRACK, 2);
    }
    switch (bits_per_samp) {
        case 8:
            ALsetsampfmt(audio_port_config, AL_SAMPFMT_TWOSCOMP);
            ALsetwidth(audio_port_config, AL_SAMPLE_8);
            AFinitsampfmt(audio_file_setup, AF_DEFAULT_TRACK,
                                                  AF_SAMPFMT_TWOSCOMP, 8);
            break;
        default:
        case 16:
            ALsetsampfmt(audio_port_config, AL_SAMPFMT_TWOSCOMP);
            ALsetwidth(audio_port_config, AL_SAMPLE_16);
            AFinitsampfmt(audio_file_setup, AF_DEFAULT_TRACK,
                                                  AF_SAMPFMT_TWOSCOMP, 16);
            break;
        case 24:
            ALsetsampfmt(audio_port_config, AL_SAMPFMT_TWOSCOMP);
            ALsetwidth(audio_port_config, AL_SAMPLE_24);
            AFinitsampfmt(audio_file_setup, AF_DEFAULT_TRACK,
                                                  AF_SAMPFMT_TWOSCOMP, 24);
            break;
    }
    AFinitrate(audio_file_setup, AF_DEFAULT_TRACK,
                                      (double)frames_per_sec);

    if ((audio_file = AFopenfd(fd , "w", audio_file_setup))
                                                      == AF_NULL_FILEHANDLE) {
        fprintf(stderr, "%s: failed to open audio file %s\n",
                   myname, filename);
	return -1;
    }

    caught_sigint = 0;

    audio_port      = ALopenport("recordaiff", "r", audio_port_config);
    record_audio(audio_port, audio_file);

    /*
     * clean up
     */
#ifdef DEBUG
    fprintf(stderr, "cleaning up\n");
#endif
    if ((n =  (int) AFclosefile(audio_file)) < 0)
    {
        fprintf(stderr, "%s: failed to update and close %s\n",
                    myname, filename);
	return -1;
    }
    AFfreefilesetup(audio_file_setup);
    ALcloseport(audio_port);
    ALfreeconfig(audio_port_config);

    return 1;
}

/*
 * R E C O R D _ A U D I O
 */
void
RecordClass::record_audio(ALport audio_port, AFfilehandle audio_file)
{
#ifdef DEBUG
    fprintf(stderr, "record_audio(...)\n");
#endif

    int nsamps;
    int nframes;
    char *buf;
    int samps_per_buf;
    int bytes_per_buf;
    int frames_per_buf;
    double framespersec;
    int bytes_per_samp;
    double secs_per_frame;
    double secs_per_buf;
    double sec_count;
    int samps_per_frame;
    int i;
    int frame_limit;
    int num_bufs;
    int leftover_samps;
    int leftover_frames;
    int leftover_bytes;
    int done = 0;
    long bits_per_samp;
    long total_samp_bytes;
    long total_frames;
    long samp_type;
    char compressionname[20];

    samps_per_frame = (int) AFgetchannels(audio_file, AF_DEFAULT_TRACK);
    AFgetsampfmt(audio_file, AF_DEFAULT_TRACK, &samp_type, &bits_per_samp);
    if (bits_per_samp <= 8) {
        bytes_per_samp = 1;
    }
    else if (bits_per_samp <= 16) {
        bytes_per_samp = 2;
    }
    else {
        bytes_per_samp = 4;
    }

    framespersec  = AFgetrate(audio_file, AF_DEFAULT_TRACK);
#ifdef DEBUG
  fprintf(stderr, "framespersec = %f\n", framespersec);
#endif

    secs_per_frame  = 1.0/framespersec;
    frames_per_buf  = (int) framespersec / 2;
    samps_per_buf   = frames_per_buf * samps_per_frame;
    bytes_per_buf   = samps_per_buf * bytes_per_samp;
    secs_per_buf    = secs_per_frame * frames_per_buf;

    buf          = (char *)malloc(bytes_per_buf);


    /*
     * we assume the file pointer now points to the beginning
     * of the audio sample section of the AIFF file
     */
    if (!time_limit_specified)
    {
       total_samp_bytes = 0;
       total_frames     = 0;
       sec_count        = 0.0;
       done             = 0;

       if (!quiet)
       {
          switch (compression)
          {
             default:
             case AF_COMPRESSION_NONE:
                    strcpy(compressionname, "");
                    break;
             case AF_COMPRESSION_G722:
                    strcpy(compressionname, "--> G.722");
                    break;
             case AF_COMPRESSION_G711_ALAW:
                    strcpy(compressionname, "--> A-law");
                    break;
             case AF_COMPRESSION_G711_ULAW:
                    strcpy(compressionname, "--> u-law");
                    break;
          }
          printf("%s: %7.1f Hz %6s %d-bit %s %s\n",
            filename,
            framespersec,
            samps_per_frame == 1 ? "mono" : "stereo",
            bits_per_samp,
            compressionname,
            filefmt == AF_FILE_AIFFC ? "AIFF-C" : "AIFF"
         );
       }

       while (!done)
       {
           ALreadsamps(audio_port, buf, samps_per_buf);
           if ((nframes = (int) AFwriteframes(audio_file, AF_DEFAULT_TRACK,
                                 buf, frames_per_buf)) != frames_per_buf)
           {
               fprintf(stderr,
                  "%s:  tried to write %d frames to %s, wrote only %d\n",
                   myname, frames_per_buf, filename, nframes);
               total_frames     += nframes;
               sec_count        += total_frames * secs_per_frame;
               total_samp_bytes += nframes * samps_per_frame * bytes_per_samp;
               done++;
           }
           else
           {
               total_frames     += frames_per_buf;
               total_samp_bytes += bytes_per_buf;
               sec_count += secs_per_buf;
           }

           if (verbose)
           {
               fprintf(stderr,"%s: %d sample frames  %6.2f secs  %d bytes\n",
                       myname, total_frames, sec_count, total_samp_bytes);
           }
           if (caught_sigint)
           {
               done++;
           }
       } /* while */
       /*
        * drain the audio data that has accumulated in the audio port
        * since the last time we checked
        */
       nsamps   = (int) ALgetfilled(audio_port);
       nframes  = nsamps / samps_per_frame;
       num_bufs = nframes / frames_per_buf;
       leftover_frames = nframes - (num_bufs * frames_per_buf);
       leftover_samps  = leftover_frames * samps_per_frame;

       if (!quiet && (num_bufs > 0))
       {
          fprintf(stderr, "%s: draining the audio input port ...\n",
                                  myname);
       }

       while (num_bufs)
       {
           ALreadsamps(audio_port, buf, samps_per_buf);
           if ((nframes = (int) AFwriteframes(audio_file, AF_DEFAULT_TRACK,
                                 buf, frames_per_buf)) != frames_per_buf)
           {
               fprintf(stderr,
                  "%s:  tried to write %d frames to %s, wrote only %d\n",
                   myname, frames_per_buf, filename, nframes);
               total_frames     += nframes;
               sec_count        += total_frames * secs_per_frame;
               total_samp_bytes += nframes * samps_per_frame * bytes_per_samp;
               num_bufs = 0;
           }
           else
           {
               total_frames     += frames_per_buf;
               total_samp_bytes += bytes_per_buf;
               sec_count += secs_per_buf;
               num_bufs--;
           }
           if (verbose)
           {
               fprintf(stderr,"%s: %d sample frames  %6.2f secs  %d bytes\n",
                       myname, total_frames, sec_count, total_samp_bytes);
           }
       }
       ALreadsamps(audio_port, buf, leftover_samps);
       if ((nframes = (int) AFwriteframes(audio_file, AF_DEFAULT_TRACK,
                                 buf, leftover_frames)) != leftover_frames)
       {
           fprintf(stderr,
                  "%s:  tried to write %d frames to %s, wrote only %d\n",
                   myname, leftover_frames, filename, nframes);
           total_frames     += nframes;
           sec_count        += total_frames * secs_per_frame;
           total_samp_bytes += nframes * samps_per_frame * bytes_per_samp;
       }
       else
       {
           total_frames     += leftover_frames;
           total_samp_bytes
                 += (leftover_frames * bytes_per_samp * samps_per_frame);
       }
       if (verbose)
       {
           sec_count += leftover_frames * secs_per_frame;
           fprintf(stderr,"%s: %d sample frames  %6.2f secs  %d bytes\n",
                       myname, total_frames, sec_count, total_samp_bytes);
       }
   } /* no time limit specified */

   else /* user has specified a time limit */
   {
       total_samp_bytes = 0;
       total_frames     = 0;
       sec_count        = 0.0;
       done             = 0;

      /*
       * translate time (in seconds) to number of sample frames
       */
      frame_limit     = (int)(time_limit * framespersec);
      num_bufs        = frame_limit / frames_per_buf;
      leftover_frames = frame_limit - (num_bufs * frames_per_buf);
      leftover_samps  = leftover_frames * samps_per_frame;
      leftover_bytes  = leftover_samps * bytes_per_samp;
      sec_count       = 0.0;
      done            = 0;

      if (!quiet)
      {
          switch (compression)
          {
             default:
             case AF_COMPRESSION_NONE:
                    strcpy(compressionname, "");
                    break;
             case AF_COMPRESSION_G722:
                    strcpy(compressionname, "G.722");
                    break;
             case AF_COMPRESSION_G711_ALAW:
                    strcpy(compressionname, "A-law");
                    break;
             case AF_COMPRESSION_G711_ULAW:
                    strcpy(compressionname, "u-law");
                    break;
          }
#ifdef DEBUG
          printf("%s: '%s' %6.3f sec %7.1f Hz %6s %d-bit %s %s\n",
            myname,
            filename,
            time_limit,
            framespersec,
            samps_per_frame == 1 ? "mono" : "stereo",
            bits_per_samp,
            compressionname,
            filefmt == AF_FILE_AIFFC ? "AIFF-C" : "AIFF"
         );
#endif
      }

      /*
       * write blocks of samples
       */
      for (i=0; (i<num_bufs) && (!done); i++)
      {
           ALreadsamps(audio_port, buf, samps_per_buf);
           if ((nframes = (int) AFwriteframes(audio_file, AF_DEFAULT_TRACK,
                              buf, frames_per_buf)) != frames_per_buf)
           {
               fprintf(stderr,
                  "%s:  tried to write %d samps to %s, actually wrote %d\n",
                   myname, samps_per_buf, filename, nsamps);

               total_frames     += nframes;
               sec_count        += total_frames * secs_per_frame;
               done++;
           }
           else
           {
               total_frames     += frames_per_buf;
               total_samp_bytes += bytes_per_buf;
               sec_count        += secs_per_buf;
           }
           if (verbose)
           {
               fprintf(stderr,"%s: %d sample frames  %6.2f secs  %d bytes\n",
                       myname, total_frames, sec_count, total_samp_bytes);
           }
           if (caught_sigint)
           {
               done++;
           }
      }
      /*
       * write the leftover samples
       */
      if (!done)
      {
          ALreadsamps(audio_port, buf, leftover_samps);
          if ((nframes = (int) AFwriteframes(audio_file, AF_DEFAULT_TRACK,
              buf, leftover_frames)) != leftover_frames)
          {
              fprintf(stderr,
                 "%s:  tried to write %d frames to %s, wrote only %d\n",
                  myname, leftover_frames, filename, nframes);
          }
          total_frames     += nframes;
          total_samp_bytes += nframes * samps_per_frame * bytes_per_samp;
          if (verbose)
          {
              sec_count += leftover_frames * secs_per_frame;
              fprintf(stderr, "%s: %d sample frames  %6.2f secs  %d bytes\n",
                    myname, total_frames, sec_count, total_samp_bytes);
          }
      }
   }
}

int RecordClass::start()
{
#ifdef DEBUG
    fprintf(stderr, "start()\n");
#endif
    fork = 1;
   /*
    *    Use the sproc(2) call to create an audio thread. The audio
    *    thread begins execution at the do_audio entry point. The
    *    second argument to sproc, PR_SALL, allows the two processes
    *    to share all attributes. See the manual page for the sproc(2)
    *    system call for more details.
    */
    audio_pid =
      sproc((void (*)(void *))&RecordClass::go, PR_SALL, (void *) this);
    /* (void (*)(void *))&RecordClass::beginRecording, PR_SALL); */
   if (audio_pid == -1)
   {
       fprintf(stderr, "unable to create audio thread...aborting.\n");
       return -1;
   }
   return 1;
}

void RecordClass::stop()
{
#ifdef DEBUG
      fprintf(stderr, "stop()\n");
#endif
//      if (!isLive) return;

      caught_sigint = 1;
#ifdef DEBUG
      fprintf(stderr, "waiting...");
#endif

//      pid_t pid;
//      int status;
//      pid = wait(&status);

      if (audio_pid != -1) {
	  union wait waitStatus;
	  (void) waitpid(audio_pid, (int *) (&waitStatus), 0);
      }
#ifdef DEBUG
      fprintf(stderr, "done.\n");
#endif
      isLive = FALSE;
}


void RecordClass::go(void *obj)
{
#ifdef DEBUG
    fprintf(stderr, "go()\n");
#endif
    RecordClass *objPtr = (RecordClass *) obj;

    objPtr->beginRecording();
#ifdef DEBUG
    fprintf(stderr, "return from beginRecording()\n");
    fprintf(stderr, "process = %d\n", objPtr->audio_pid);
#endif
}

short RecordClass::getIsLive()
{
    return isLive;
}

/*
 * application-defined silent error reporting routine
 * for Audio Library
 */
static void
quietALerror(long code, const char *desc, ...)
{
}

/*
 * application-defined silent error reporting routine
 * for Audio File Library
 */
static void
quietAFerror(long code, const char *desc)
{
}
